﻿/* Мною, Константином aka KilkennyCat, 05 июля 2020 года создано сиё 
 * на основе "Описание формата TopoR PCB версия 1.2.0 Апрель 2017 г.".
 * k@kilkennycat.pro
 * http://kilkennycat.ru  http://kilkennycat.pro
 */

using System;
using System.Collections.Generic;
using System.Xml.Serialization;

namespace TopoR_PCB_Classes
{
    /// <summary>
    /// Компоненты на плате (обязательный раздел).
    /// </summary>
    public class ComponentsOnBoard
    {
        /// <summary>
        /// Описание компонента на плате.
        /// </summary>
        /// <remarks>! Если компонент расположен на нижней стороне платы, его посадочное место отображается зеркально относительно вертикальной оси посадочного места, описанного в библиотеке(т.е.без угла поворота). Стеки контактных площадок переворачиваются.</remarks>
        public class CompInstance
        {

            /// <summary>
            /// Описание контакта компонента на плате.
            /// </summary>
            /// <remarks>! Если PadstackRef не указан, то стек контактных площадок берётся из посадочного места.</remarks>
            public class CompInstance_Pin
            {
                /// <summary>
                /// Номер контакта компонента.
                /// </summary>
                [XmlAttribute("padNum", DataType = "int")]
                public int _padNum;

                /// <summary>
                /// Ссылка на стек контактных площадок.
                /// </summary>
                [XmlElement("PadstackRef")]
                public PadstackRef _PadstackRef;
                public bool ShouldSerialize_PadstackRef()
                {
                    return _PadstackRef != null;
                }
                /// <summary>
                /// Точка привязки объекта.
                /// </summary>
                [XmlElement("Org")]
                public Org _Org;
            }

            /// <summary>
            /// Описание монтажного отверстия в компоненте на плате.
            /// </summary>
            public class CompInstance_Mnthole
            {
                /// <summary>
                /// Ссылка на монтажное отверстие в посадочном месте.
                /// </summary>
                [XmlAttribute("mntholeRef")]
                public string _mntholeRef;

                /// <summary>
                /// Задаёт угол в градусах c точностью до тысячных долей.
                /// </summary>
                [XmlAttribute("angle", DataType = "float")]
                public float _angle;

                /// <summary>
                /// Ссылка на стек контактных площадок.
                /// </summary>
                [XmlElement("PadstackRef")]
                public PadstackRef _PadstackRef;

                /// <summary>
                /// Cсылка на цепь.
                /// </summary>
                [XmlElement("NetRef")]
                public NetRef _NetRef;
            }

            /// <summary>
            /// Описание атрибута компонента на плате.
            /// </summary>
            public class CompInstance_Attribute
            {
                /// <summary>
                /// Описание ярлыка компонента на плате.
                /// </summary>
                public class CompInstance_Attribute_Label
                {
                    /// <summary>
                    /// Задаёт угол в градусах c точностью до тысячных долей.
                    /// </summary>
                    [XmlAttribute("angle", DataType = "float")]
                    public float _angle;

                    /// <summary>
                    /// Параметр надписей и ярлыков: зеркальность отображения.
                    /// </summary>
                    [XmlAttribute("mirror")]
                    public Bool _mirror;
                    [XmlIgnore]
                    public bool _mirrorSpecified { get { return _mirror != Bool.off; } }

                    /// <summary>
                    /// Параметр надписей (ярлыков): способ выравнивания текста.
                    /// </summary>
                    [XmlAttribute("align")]
                    public align _align;

                    /// <summary>
                    /// Флаг видимости.
                    /// </summary>
                    [XmlAttribute("visible")]
                    public Bool _visible;
                    [XmlIgnore]
                    public bool _visibleSpecified { get { return _visible != Bool.off; } }

                    /// <summary>
                    /// Ссылка на слой.
                    /// </summary>
                    [XmlElement("LayerRef")]
                    public LayerRef _LayerRef;

                    /// <summary>
                    /// Ссылка на стиль надписей.
                    /// </summary>
                    [XmlElement("TextStyleRef")]
                    public TextStyleRef _TextStyleRef;

                    /// <summary>
                    /// Точка привязки объекта.
                    /// </summary>
                    [XmlElement("Org")]
                    public Org _Org;
                }

                /// <summary>
                /// Тип предопределённого атрибута компонента.
                /// </summary>
                [XmlAttribute("type")]
                public type _type;

                /// <summary>
                /// Имя объекта или ссылка на именованный объект.
                /// </summary>
                [XmlAttribute("name")]
                public string _name;

                /// <summary>
                /// Значение атрибута.
                /// </summary>
                [XmlAttribute("value")]
                public string _value;

                /// <summary>
                /// Ярлыки.
                /// </summary>
                [XmlElement("Label")]
                public List<CompInstance_Attribute_Label> _Labels;
                public bool ShouldSerialize_Labels()
                {
                    return _Labels?.Count > 0;
                }
            }

            /// <summary>
            /// Имя объекта или ссылка на именованный объект.
            /// </summary>
            [XmlAttribute("name")]
            public string _name;

            /// <summary>
            /// Уникальный идентификатор компонента. Используется при синхронизации. Необязательный атрибут.
            /// Если не задан, то будет создан при импорте файла.
            /// </summary>
            [XmlAttribute("uniqueId")]
            public string _uniqueId;

            /// <summary>
            /// Сторона объекта.
            /// </summary>
            /// <remarks>!Значение Both возможно только при описании запретов размещения.</remarks>
            [XmlAttribute("side")]
            public side _side;

            /// <summary>
            /// Задаёт угол в градусах c точностью до тысячных долей.
            /// </summary>
            [XmlAttribute("angle", DataType = "float")]
            public float _angle;

            /// <summary>
            /// Признак фиксации.
            /// </summary>
            [XmlAttribute("fixed")]
            public Bool _fixed;
            [XmlIgnore]
            public bool _fixedSpecified { get { return _fixed != Bool.off; } }

            /// <summary>
            /// Ссылка на схемный компонент.
            /// </summary>
            [XmlElement("ComponentRef")]
            public ComponentRef _ComponentRef;

            /// <summary>
            /// Ссылка на посадочное место.
            /// </summary>
            [XmlElement("FootprintRef")]
            public FootprintRef _FootprintRef;

            /// <summary>
            /// Точка привязки объекта.
            /// </summary>
            [XmlElement("Org")]
            public Org _Org;

            /// <summary>
            /// Контакты компонента на плате.
            /// </summary>
            [XmlArray("Pins")]
            [XmlArrayItem("Pin")]
            public List<CompInstance_Pin> _Pins;
            public bool ShouldSerialize_Pins()
            {
                return _Pins?.Count > 0;
            }
            /// <summary>
            /// Монтажные отверстия.
            /// </summary>
            [XmlArray("Mntholes")]
            [XmlArrayItem("Mnthole")]
            public List<CompInstance_Mnthole> _Mntholes;
            public bool ShouldSerialize_Mntholes()
            {
                return _Mntholes?.Count > 0;
            }
            /// <summary>
            /// Атрибуты компонента.
            /// </summary>
            [XmlArray("Attributes")]
            [XmlArrayItem("Attribute")]
            public List<CompInstance_Attribute> _Attributes;
            public bool ShouldSerialize_Attributes()
            {
                return _Attributes?.Count > 0;
            }


            /************************************************************************
             * Здесь находятся функции для работы с элементами класса CompInstance. *
             * Они не являются частью формата TopoR PCB.                            *
             * **********************************************************************/
            /// <summary>
            /// Для отображения имени компонента
            /// </summary>
            /// <returns></returns>
            public override string ToString()
            {
                return _name;
            }
            /***********************************************************************/
        }

        /// <summary>
        /// Описание одиночного контакта..
        /// </summary>
        public class FreePad
        {
            /// <summary>
            /// Сторона объекта.
            /// </summary>
            [XmlAttribute("side")]
            public side _side;

            /// <summary>
            /// Задаёт угол в градусах c точностью до тысячных долей.
            /// </summary>
            [XmlAttribute("angle", DataType = "float")]
            public float _angle;

            /// <summary>
            /// Признак фиксации.
            /// </summary>
            [XmlAttribute("fixed")]
            public Bool _fixed;
            [XmlIgnore]
            public bool _fixedSpecified { get { return _fixed != Bool.off; } }

            /// <summary>
            /// Ссылка на стек контактных площадок.
            /// </summary>
            [XmlElement("PadstackRef")]
            public PadstackRef _PadstackRef;

            /// <summary>
            /// Cсылка на цепь.
            /// </summary>
            [XmlElement("NetRef")]
            public NetRef _NetRef;

            /// <summary>
            /// Точка привязки объекта.
            /// </summary>
            [XmlElement("Org")]
            public Org _Org;
        }

        /// <summary>
        /// Версия раздела.
        /// </summary>
        [XmlAttribute("version")]
        public string _version;

        /// <summary>
        /// Описание компонентов на плате (инстанции компонентов)
        /// </summary>
        [XmlArray("Components")]
        [XmlArrayItem("CompInstance")]
        public List<CompInstance> _Components;
        public bool ShouldSerialize_Components()
        {
            return _Components?.Count > 0;
        }
        /// <summary>
        /// Описание одиночных контактов.(инстанции компонентов)
        /// </summary>
        [XmlArray("FreePads")]
        [XmlArrayItem("FreePad")]
        public List<FreePad> _FreePads;
        public bool ShouldSerialize_FreePads()
        {
            return _FreePads?.Count > 0;
        }


        /*****************************************************************************
         * Здесь находятся функции для работы с элементами класса ComponentsOnBoard. *
         * Они не являются частью формата TopoR PCB.                                 *
         * ***************************************************************************/
        
        /// <summary>
        /// Индекс компонента
        /// </summary>
        /// <param name="name">уникальное имя компонента</param>
        /// <returns>индекс компонента или -1, если компонент отсутствует</returns>
        public int ComponentIndexOf(string name)
        {
            for (int x = _Components?.Count ?? 0; x > 0; x--)
            {
                if (_Components[x - 1]._name.ToUpper() == name.ToUpper())
                {
                    return x - 1;
                }
            }
            return -1;
        }

        /// <summary>
        /// Добавление компонента
        /// </summary>
        /// <param name="name">Имя нового компонента. Если имя неуникально, будет добавлен префикс _</param>
        /// <param name="dist">текущие единицы измерения</param>
        /// <param name="package">упаковка</param>
        public void AddComponent(string name, dist dist, LocalLibrary.Package package)
        {
            float x = 0, y = 0;//координаты нового компонента
            if (_Components == null)
            {
                return;
            }
            while (ComponentIndexOf(name) >= 0) //проверка на уникальность имени и добавление префикса
            {
                name += "_";
            }
            for (int i = _Components.Count; i > 0; i--)//вычисление максимально возможных координат
            {
                x = Math.Max(x, _Components[i - 1]._Org._x);
                y = Math.Max(y, _Components[i - 1]._Org._y);
            }
            float offset = Ut.UnitsConvert(3, dist, dist.mm);
            x += offset;//добавление небольшого смещения 
            y += offset;

            _Components.Add(
                new CompInstance
                {
                    _name = name,
                    _side = side.Top,
                    _uniqueId = UniqueId(),
                    _angle = 0,
                    _fixed = Bool.off,
                    _Org = new Org { _x = x, _y = y },
                     _ComponentRef = new ComponentRef { _ReferenceName = package._ComponentRef._ReferenceName },
                     _FootprintRef = new FootprintRef { _ReferenceName = package._FootprintRef._ReferenceName }
                }
           );
 
        }
        /// <summary>
        /// Удаление компонента по имени
        /// </summary>
        /// <param name="name">уникальный имя компонента</param>
        /// <returns>true - если было произведено удаление, иначе (компонент не найден) - false</returns>
        public bool RemoveComponent(string name)
        {
            int x = ComponentIndexOf(name);
            if (x >= 0)
            {
                _Components.RemoveAt(x);
                return true;
            }
            return false;
        }        

        /// <summary>
        /// Переименование компонента
        /// </summary>
        /// <param name="oldname">старое имя компонента</param>
        /// <param name="newname">новое имя компонента</param>
        /// <returns>индекс компонента, если было произведено переименование, -1, если компонент не найден</returns>
        public int RenameComponent(string oldname, string newname)
        {
            if (newname == "")
            {
                return -1;
            }
            int x = ComponentIndexOf(oldname);
            if (x >= 0)
            {
                _Components[x]._name = newname;
                return x;
            }
            return -1;
        }

        /// <summary>
        /// Генерация уникального идентификатора
        /// </summary>
        /// <returns>string like "ABCDEFGH"</returns>
        public string UniqueId()
        {
            string ABC = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
            string uniqueId = "";
            Random rnd = new Random();
            for (int i = 0; i < 8; i++)
            {
                uniqueId += ABC[rnd.Next(26)];
            }
            if (_Components?.Count > 0)
            {
                foreach (CompInstance c in _Components)
                {
                    if (c._uniqueId == uniqueId)
                    {
                        uniqueId = UniqueId();
                    }
                }
            }
            return uniqueId;
        }

        /*************************************************************************************/

    }
}
